import numpy as np
import pandas as pd
import seaborn as sns
import warnings

from sklearn.metrics import classification_report, confusion_matrix

from sklearn.preprocessing import StandardScaler, LabelEncoder
from sklearn.tree import DecisionTreeClassifier

from utilities import plot_counts
from utilities import plot_features, ordinal_encode, normalize, plot_confusion_matrix, print_high_confidence_samples, import_data

warnings.filterwarnings("ignore")


# Constants
y_data = 'class'
df_train, df_test, model_name = import_data(
    train_file_path = "nsl-kdd-dataset/" + "KDDTrain+.arff",
    test_file_path =  "nsl-kdd-dataset/" + "KDDTest+.arff",
    model_name = "Decision Tree")
sc = StandardScaler()
enc = LabelEncoder()

is_threat = df_train[y_data].unique()
if len(is_threat) != 2:
    raise Exception("Target must be a binary decision.")

# Normalize data
ordinal_encode(df = df_train, categories = is_threat, target = y_data)
ordinal_encode(df = df_test, categories = is_threat, target = y_data)

normalize(df_train, df_test, y_data, sc, enc)

# Plot absolute quantities of class 0 and class 1
sns.countplot(x = y_data, data = df_train)
plot_counts(model_name = model_name)

# Separate X and y
X_train = df_train.select_dtypes(include=[np.number]).drop(columns = [y_data])
X_test = df_test.select_dtypes(include=[np.number]).drop(columns = [y_data])
y_train = df_train[[y_data]]
y_test = df_test[[y_data]]

# Different parameters
# Criterion modifiers ('criterion'):
# 'gini' is computationally faster because it does not involve computing logarithms.
# 'entropy' tends to work better when class distributions are highly imbalanced.
# 'log_loss' is more useful when you care about predicting class probabilities accurately.

# Cost Complexity Pruning ('ccp_alpha'):
# The ccp_alpha parameter controls the pruning process in a decision tree.
# Pruning is the process of removing parts of the tree that do not provide significant improvement in prediction,
# to reduce overfitting.
# Higher ccp_alpha -> more pruning -> simpler tree and vice versa

# Max Depth ('max_depth'):
# Limits how much a tree can grow
# Small depth -> simpler, less overfitting and vice versa

# Minimum Samples Fit ('min_samples_split'):
# Determines how the min. of samples to split a node
# Higher value -> lesser splits -> simpler tree and vice versa

# Random state ('random_state'):
# Allows reproducibility

# What is overfitting?
# Machine learning model "memorizes" the training data, rather than understanding the underlying pattern

# Training model
dtc = DecisionTreeClassifier()
dtc.fit(X_train, y_train)
y_prediction = dtc.predict(X_test)
print("Classification report: \n", classification_report(y_test, y_prediction))
plot_confusion_matrix(confusion_matrix = confusion_matrix(y_test, y_prediction),
                      accuracy = dtc.score(X_test, y_test),
                      model_name=model_name)

# Determine feature importance
features = pd.DataFrame(dtc.feature_importances_,
                        index= X_train.columns,
                        columns=['Importance']).sort_values(by='Importance', ascending=False)
plot_features(features, model_name = model_name)

def predict(prediction_input):
    if len(prediction_input) == 0:
        return
    input_data = pd.DataFrame(prediction_input, columns = X_train.columns)
    return dtc.predict(input_data)

print_high_confidence_samples(model = dtc, x = X_train)
