import numpy as np
import pandas as pd
import seaborn as sns

import warnings
from sklearn.preprocessing import StandardScaler, LabelEncoder

from utilities import ordinal_encode, heat_map, plot_features, plot_confusion_matrix, normalize, \
    print_high_confidence_samples, plot_counts, import_data, plot_roc_curve

warnings.filterwarnings("ignore")

from sklearn.linear_model import LogisticRegression
from sklearn.metrics import classification_report, confusion_matrix

# Constants

y_data = 'class'
df_train, df_test, model_name = import_data(
    train_file_path = "nsl-kdd-dataset/" + "KDDTrain+.arff",
    test_file_path =  "nsl-kdd-dataset/" + "KDDTest+.arff",
    model_name = "Logistic Regression")
sc = StandardScaler()
enc = LabelEncoder()

# Data inspection and pre-processing

is_threat = df_train[y_data].unique()
if len(is_threat) != 2:
    raise Exception("Logistic Regression only works for binary classification.")

numerical_columns = df_train.select_dtypes(include = np.number).columns
label_columns = df_train.select_dtypes(include=object, exclude = np.number).columns

# Normalize data
ordinal_encode(df = df_train, categories = is_threat, target = y_data)
ordinal_encode(df = df_test, categories = is_threat, target = y_data)

normalize(df_train, df_test, y_data, sc, enc)

# Correlation
heat_map(df_train, model_name = model_name)

# Count plot
sns.countplot(x = y_data, data = df_train)
plot_counts(model_name=model_name)

# Separate X and y
X_train = df_train.select_dtypes(include=[np.number]).drop(columns = [y_data])
X_test = df_test.select_dtypes(include=[np.number]).drop(columns = [y_data])
y_train = df_train[[y_data]]
y_test = df_test[[y_data]]

# 'penalty'
# Helps preventing overfitting.
# l2 (default), l1, elasticnet (l1 + l2), none (not recommended)

# 'C'
# Variable that determines "how much is the model allowed to fit to the data?"
# smaller C -> simpler model (less overfitting) and vice versa

# 'solver'
# Algorithms that optimize logistic regression
# 'lbfgs' (default) and works well with l2
# 'liblinear' good for small datasets and works well with l1
# 'saga' handles larger datasets, supports all penalty types
# 'newton-cg', 'sag' useful for larger datasets

# 'l1-ratio'
# Support utility for 'elasticnet', determines the ratio of l1 and l2 regularization

# 'max_iter'
# Maximum of iterations for the solver, sometimes needed if the data is large and the model struggles to converge
# Ergo larger nsl-kdd-dataset or elasticnet may require a larger 'max_iter'

# 'class_weight'
# Handles class imbalance by giving unrepresented classes higher weights. To enable use 'balanced'

# 'random_state'
# Allows reproducibility

model = LogisticRegression()
model.fit(X_train, y_train)
y_prediction = model.predict(X_test)
plot_confusion_matrix(confusion_matrix = confusion_matrix(y_test, y_prediction),
                      accuracy = model.score(X_test, y_test),
                      model_name=model_name)
print("Classification report: \n", classification_report(y_test, y_prediction))

# Display feature importance
features = pd.DataFrame(model.coef_[0], index= X_train.columns, columns=['Importance']).sort_values(by='Importance', ascending=False)
plot_features(features = features,
              info_text= "Positive coefficient = More probable to land in class 1 \n"
                         "Negative coefficient = Less probable to land in class 1",
              model_name=model_name)

# Display samples for which the model is 90% confident
print_high_confidence_samples(model, X_train)

# Calculate prediction probabilities for ROC curve
y_score = model.predict_proba(X_test)[:, 1]

# Plot ROC curve using the function from utilities
plot_roc_curve(y_test, y_score, model_name=model_name)

def predict(prediction_input):
    if len(prediction_input) == 0:
        return
    input_df = pd.DataFrame(prediction_input, columns = X_train.columns)
    input_df[numerical_columns] = sc.transform(input_df[numerical_columns])
    return ["anomaly" if x == 1 else "normal" for x in model.predict(input_df)]


